@@ -3,9 +3,9 @@
  *
  * Implementation of the Domain-Free Mandatory Access Control.
  *
- * Copyright (C) 2005-2006  NTT DATA CORPORATION
+ * Copyright (C) 2005-2007  NTT DATA CORPORATION
  *
- * Version: 1.2   2006/09/03
+ * Version: 1.4.3   2007/11/11
  *
  * This file is applicable to both 2.4.30 and 2.6.11 and later.
  * See README.ccs for ChangeLog.
@@ -16,7 +16,12 @@
 #include <linux/ccs_common.h>
 #include <linux/sakura.h>
 #include <linux/realpath.h>
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,20)
+#include <linux/mount.h>
+#include <linux/mnt_namespace.h>
+#else
 #include <linux/namespace.h>
+#endif
 #if LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0)
 #include <linux/namei.h>
 #endif
@@ -26,7 +31,13 @@
 int SAKURA_MayMount(struct nameidata *nd)
 {
 	int flag = 0;
+#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,19)
 	struct namespace *namespace = current->namespace;
+#elif LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20)
+	struct namespace *namespace = current->nsproxy->namespace;
+#else
+	struct mnt_namespace *namespace = current->nsproxy->mnt_ns;
+#endif
 	if (!CheckCCSFlags(CCS_SAKURA_DENY_CONCEAL_MOUNT)) return 0;
 	if (namespace) {
 		struct list_head *p;
@@ -67,21 +78,18 @@ int SAKURA_MayMount(struct nameidata *nd)
 	if (flag) {
 		int error = -EPERM;
 		const int is_enforce = CheckCCSEnforce(CCS_SAKURA_DENY_CONCEAL_MOUNT);
-		char *page = ccs_alloc(PAGE_SIZE);
-		if (page) {
-			if (realpath_from_dentry(nd->dentry, nd->mnt, page, PAGE_SIZE - 1) == 0) {
-				const char *exename = GetEXE();
-				printk("SAKURA-%s: mount %s (pid=%d:exe=%s): Permission denied.\n", GetMSG(is_enforce), page, current->pid, exename);
-				if (is_enforce && CheckSupervisor("# %s is requesting\nmount on %s\n", exename, page) == 0) error = 0;
-				ccs_free(exename);
-			}
-			ccs_free(page);
+		const char *dir = realpath_from_dentry(nd->dentry, nd->mnt);
+		if (dir) {
+			const char *exename = GetEXE();
+			printk("SAKURA-%s: mount %s (pid=%d:exe=%s): Permission denied.\n", GetMSG(is_enforce), dir, current->pid, exename);
+			if (is_enforce && CheckSupervisor("# %s is requesting\nmount on %s\n", exename, dir) == 0) error = 0;
+			ccs_free(exename);
 		}
+		ccs_free(dir);
 		if (is_enforce) return error;
 	}
 	return 0;
 }
-
 EXPORT_SYMBOL(SAKURA_MayMount);
 
 /***** SAKURA Linux end. *****/
